@@ -23,6 +23,8 @@ from lmnr.opentelemetry_lib.tracing.attributes import (
 )
 from lmnr.opentelemetry_lib import MAX_MANUAL_SPAN_PAYLOAD_SIZE
 from lmnr.opentelemetry_lib.decorators import json_dumps
+from lmnr.sdk.utils import get_otel_env_var
+
 from opentelemetry import trace
 from opentelemetry import context as context_api
 from opentelemetry.trace import INVALID_TRACE_ID, Span, Status, StatusCode, use_span
@@ -140,7 +142,12 @@ class Laminar:
             return
 
         cls.__project_api_key = project_api_key or from_env("LMNR_PROJECT_API_KEY")
-        if not cls.__project_api_key:
+
+        if (
+            not cls.__project_api_key
+            and not get_otel_env_var("ENDPOINT")
+            and not get_otel_env_var("HEADERS")
+        ):
             raise ValueError(
                 "Please initialize the Laminar object with"
                 " your project API key or set the LMNR_PROJECT_API_KEY"
@@ -149,14 +156,15 @@ class Laminar:
 
         cls._initialize_logger()
 
-        url = base_url or from_env("LMNR_BASE_URL") or "https://api.lmnr.ai"
-        url = url.rstrip("/")
-        if not url.startswith("http:") and not url.startswith("https:"):
-            url = f"https://{url}"
-        if match := re.search(r":(\d{1,5})$", url):
-            url = url[: -len(match.group(0))]
-            cls.__logger.info(f"Ignoring port in base URL: {match.group(1)}")
-        http_url = base_http_url or url
+        url = base_url or from_env("LMNR_BASE_URL")
+        if url:
+            url = url.rstrip("/")
+            if not url.startswith("http:") and not url.startswith("https:"):
+                url = f"https://{url}"
+            if match := re.search(r":(\d{1,5})$", url):
+                url = url[: -len(match.group(0))]
+                cls.__logger.info(f"Ignoring port in base URL: {match.group(1)}")
+        http_url = base_http_url or url or "https://api.lmnr.ai"
         if not http_url.startswith("http:") and not http_url.startswith("https:"):
             http_url = f"https://{http_url}"
         if match := re.search(r":(\d{1,5})$", http_url):
@@ -421,14 +429,14 @@ class Laminar:
 
         Usage example:
         ```python
-        from src.lmnr import Laminar, use_span
+        from src.lmnr import Laminar
         def foo(span):
-            with use_span(span):
+            with Laminar.use_span(span):
                 with Laminar.start_as_current_span("foo_inner"):
                     some_function()
         
         def bar():
-            with use_span(span):
+            with Laminar.use_span(span):
                 openai_client.chat.completions.create()
         
         span = Laminar.start_span("outer")
@@ -773,6 +781,20 @@ class Laminar:
             return False
         return TracerManager.flush()
 
+    @classmethod
+    def force_flush(cls):
+        """Force flush the internal tracer.
+
+        Actually shuts down the span processor and re-initializes it as long
+        as it is a LaminarSpanProcessor. This is not recommended in production
+        workflows, but is useful at the end of Lambda functions, where a regular
+        flush might be killed by the Lambda runtime, because the actual export
+        inside it runs in a background thread.
+        """
+        if not cls.is_initialized():
+            return
+        TracerManager.force_reinit_processor()
+
     @classmethod
     def shutdown(cls):
         if cls.is_initialized():
